//
// Copyright (C) 2013 by Autodesk, Inc. All Rights Reserved.
//
// The information contained herein is confidential, proprietary
// to Autodesk, Inc.,  and considered a trade  secret as defined
// in section 499C of the penal code of the State of California.
// Use  of  this  information  by  anyone other  than authorized
// employees of Autodesk, Inc.  is granted  only under a written
// non-disclosure  agreement,  expressly  prescribing  the scope
// and manner of such use.
//
// AUTODESK  MAKES  NO  WARRANTIES,  EXPRESS  OR  IMPLIED, AS  TO THE
// CORRECTNESS OF THIS CODE OR ANY DERIVATIVE WORKS WHICH INCORPORATE
// IT. AUTODESK PROVIDES THE CODE ON AN  "AS-IS" BASIS AND EXPLICITLY
// DISCLAIMS ANY  LIABILITY, INCLUDING  CONSEQUENTIAL AND  INCIDENTAL
// DAMAGES  FOR ERRORS,  OMISSIONS,  AND OTHER  PROBLEMS IN THE CODE.
//
// Use, duplication, or disclosure by the U.S. Government is subject
// to restrictions set forth in  FAR 52.227-19  (Commercial Computer
// Software  Restricted  Rights)  and  DFAR   252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software),  as applicable.
//
// @author Valentin Koch <valentin.koch@autodesk.com>
 //


#ifdef _DEBUG
#include <ctime>
#endif
#include <limits>

#include "cbcsolver.h"

#include "CoinModel.hpp"
#include "CbcModel.hpp"
#include "OsiClpSolverInterface.hpp"

#include "stdlib.h"

//#define _CBC_DEBUG 1

aop::MipSolver* aop::create_cbc_solver()
{
	return new aop::CbcSolver();
}


aop::CbcSolver::CbcSolver()
	:num_vars_(0),
	 num_rows_(0),
	 coin_model_(new CoinModel()),
	 max_time_(std::numeric_limits<double>::infinity())
{
}


aop::CbcSolver::~CbcSolver()
{
}


void aop::CbcSolver::clear_model()
{
	coin_model_.reset(new CoinModel());
	num_vars_ = 0;
	num_rows_ = 0;
}


void aop::CbcSolver::add_var_continuous(const double lb, const double ub, const double cost, const std::string& name)
{
	coin_model_->setObjective(num_vars_,cost);
	coin_model_->setColumnBounds(num_vars_,
				     (lb == - LpSolver::INFTY ? -COIN_DBL_MAX : lb),
				     (ub == LpSolver::INFTY ? COIN_DBL_MAX : ub));
	if (!name.empty())
		coin_model_->setColumnName(num_vars_,name.c_str());
	++num_vars_;
	return;
}


void aop::CbcSolver::add_var_integer(const double lb, const double ub, const double cost, const std::string& name)
{
	int tmp = num_vars_;
	add_var_continuous(lb,ub,cost,name);
	coin_model_->setInteger(tmp);
	return;
}


void aop::CbcSolver::add_var_binary(const double cost, const std::string& name)
{
	add_var_integer(0,1,cost,name);
	return;
}


void aop::CbcSolver::add_constraint(const std::vector<int>& vars, const std::vector<double>& coef,
			       const LpSolver::Sign sign, const double rhs, const std::string& name)
{
	if (vars.size() != coef.size())
		throw std::invalid_argument(_DNT("Vector of variables needs same size as vector of coefficients."));
	if (vars.empty()) 
		throw std::invalid_argument(_DNT("Trying to add a constraint with empty left hand side."));

	switch (sign) {
	case LpSolver::LESS_EQUAL:
		coin_model_->addRow(vars.size(), &vars[0], &coef[0], -COIN_DBL_MAX, rhs);
		break;
	case LpSolver::EQUAL:
		coin_model_->addRow(vars.size(), &vars[0], &coef[0], rhs, rhs);
		break;
	case LpSolver::GREATER_EQUAL:
		coin_model_->addRow(vars.size(), &vars[0], &coef[0], rhs, COIN_DBL_MAX);
		break;
	}
	
	if (!name.empty())
		coin_model_->setRowName(num_rows_,name.c_str());

	++num_rows_;
	return;
}


void aop::CbcSolver::set_model_sense(const LpSolver::ModelSense sense)
{
	switch (sense) {
	case LpSolver::MINIMIZE:
		coin_model_->setOptimizationDirection(1);
		break;
	case LpSolver::MAXIMIZE:
		coin_model_->setOptimizationDirection(-1);
		break;
	}
}


void aop::CbcSolver::set_maximum_solving_time(const double seconds)
{
	max_time_ = seconds;
}


aop::LpSolver::Status aop::CbcSolver::solve(double& value, std::vector<double>& solution)
{
#ifdef _CBC_DEBUG
	char buf[80];
	const std::time_t curt(std::time(NULL));
	std::strftime(buf,80,"%Y%m%d-%H%M%S.mps",std::localtime(&curt));
	coin_model_->writeMps(buf);
#endif

	// See Cbc examples for explanation of set-up.a
	OsiClpSolverInterface clp_solver;
	clp_solver.loadFromCoinModel(*(coin_model_.get()));

	CbcModel cbc_model(clp_solver);

#ifdef _CBC_DEBUG
#ifdef _GTEST
	const int CBC_LOGLEVEL = 0;
#else
	const int CBC_LOGLEVEL = 1;
#endif
#else
	const int CBC_LOGLEVEL = 0;
#endif
	cbc_model.setLogLevel(CBC_LOGLEVEL);
	if (max_time_ != std::numeric_limits<double>::infinity())
		cbc_model.setMaximumSeconds(max_time_);

	cbc_model.branchAndBound();

	value = cbc_model.getObjValue();
	int num_cols = clp_solver.getNumCols();
	const double* data = cbc_model.solver()->getColSolution();

	solution.clear();
	solution.reserve(num_cols);
	solution.insert(solution.begin(), data, data + num_cols);
	
	if (cbc_model.isProvenOptimal())
		return LpSolver::OPTIMAL;
	else if (cbc_model.isProvenInfeasible() || cbc_model.isProvenDualInfeasible()
		 || cbc_model.isInitialSolveProvenPrimalInfeasible()
		 || cbc_model.isInitialSolveProvenDualInfeasible()) 
		return LpSolver::INFEASIBLE;
	else if (cbc_model.isContinuousUnbounded()) 
		return LpSolver::UNBOUNDED;
	else if (cbc_model.isNodeLimitReached() || cbc_model.isSecondsLimitReached()
		   || cbc_model.isSolutionLimitReached())
		return LpSolver::TIMEOUT;
	else if (cbc_model.isAbandoned() || cbc_model.isInitialSolveAbandoned())
		return LpSolver::NUMERIC;
   
	return LpSolver::UNKNOWN;

}
