/** 
 * @file cbcsolver.h
 *
 * This file contains a the class Autodesk::CbcSolver which is a wrapper around CbcSolver API calls.
 *
 * Copyright (C) 2013 by Autodesk, Inc. All Rights Reserved.
 *
 * The information contained herein is confidential, proprietary
 * to Autodesk, Inc.,  and considered a trade  secret as defined
 * in section 499C of the penal code of the State of California.
 * Use  of  this  information  by  anyone other  than authorized
 * employees of Autodesk, Inc.  is granted  only under a written
 * non-disclosure  agreement,  expressly  prescribing  the scope
 * and manner of such use.
 *
 * AUTODESK  MAKES  NO  WARRANTIES,  EXPRESS  OR  IMPLIED, AS  TO THE
 * CORRECTNESS OF THIS CODE OR ANY DERIVATIVE WORKS WHICH INCORPORATE
 * IT. AUTODESK PROVIDES THE CODE ON AN  "AS-IS" BASIS AND EXPLICITLY
 * DISCLAIMS ANY  LIABILITY, INCLUDING  CONSEQUENTIAL AND  INCIDENTAL
 * DAMAGES  FOR ERRORS,  OMISSIONS,  AND OTHER  PROBLEMS IN THE CODE.
 *
 * Use, duplication, or disclosure by the U.S. Government is subject
 * to restrictions set forth in  FAR 52.227-19  (Commercial Computer
 * Software  Restricted  Rights)  and  DFAR   252.227-7013(c)(1)(ii)
 * (Rights in Technical Data and Computer Software),  as applicable.
 *
 * @author: Valentin Koch <valentin.koch@autodesk.com>
 */
#ifndef CBC_SOLVER_H_
#define CBC_SOLVER_H_

#ifdef _WIN32
#ifdef CBC_DLL
#define CBC_API __declspec(dllexport)
#else
#define CBC_API __declspec(dllimport)
#endif
#else
#define CBC_API
#endif


#include <string>
#include <vector>
#include <memory>

#include "mipsolver.h"

class CoinModel;

namespace aop
{

#ifdef _WIN32
/**
 * Factory call to produce a CbcSolver for Windows DLL
 */
	CBC_API MipSolver* create_cbc_solver();
#endif

/**
 * @class CbcSolver
 *
 * Implementation of abstract MilpSolver interface. This class is a wrapper around the Cbc/Clp API from the Coin-OR
 * project. For more information about Cbc/Clp, we refer to
 *
 * @cite http://projects.coin-or.org/Cbc
 */
class CbcSolver : public MipSolver
{
public:
	/// Initializes a new, empty CoinModel object.
	CbcSolver();
	/// Destructor.
	virtual ~CbcSolver();

	/// Deletes current CoinModel object and creates a new, empty one.
	virtual void clear_model();

	/// Add continuous variable with lower bound lb, upper bound ub, and cost cost.	
	virtual void add_var_continuous(const double lb, const double ub, const double cost, const std::string& name = "");

	/// Add integer variable with lower bound lb, upper bound ub, and cost cost.
	void add_var_integer(const double lb, const double ub, const double cost, const std::string& name = "");

	/// Add binary variable with lower bound lb, upper bound ub, and cost cost.
	void add_var_binary(const double cost, const std::string& name = "");

	/** 
	 * Add constraint with left hand side term given by variable positions in vars
	 * and corresponding coefficients in coef, sign, and right hand side constant rhs
	 */
	virtual void add_constraint(const std::vector<int>& vars, const std::vector<double>& coef,
				    const Sign sign, const double rhs, const std::string& name = "");

	/// Minimize or maximize (default for a model is minimize).
	virtual void set_model_sense(const ModelSense sense);

	/// Set the maximum solving time in seconds.
	virtual void set_maximum_solving_time(const double seconds);

	/// Solve the current model, return status after solve and fills current solution value and solution vector.
	virtual Status solve(double& value, std::vector<double>& solution);

private:
	unsigned int num_vars_;
	unsigned int num_rows_;
	std::shared_ptr<CoinModel> coin_model_;
	double max_time_;
};


} // namespace aop

#endif
