/**
 * @file lpsolver.h
 *
 * This file contains the declaration of the abstract class LpSolver and its 
 * corresponding runtime exception solver_error.
 *
 * Copyright (C) 2013 by Autodesk, Inc. All Rights Reserved.
 *
 * The information contained herein is confidential, proprietary
 * to Autodesk, Inc.,  and considered a trade  secret as defined
 * in section 499C of the penal code of the State of California.
 * Use  of  this  information  by  anyone other  than authorized
 * employees of Autodesk, Inc.  is granted  only under a written
 * non-disclosure  agreement,  expressly  prescribing  the scope
 * and manner of such use.
 *
 * AUTODESK  MAKES  NO  WARRANTIES,  EXPRESS  OR  IMPLIED, AS  TO THE
 * CORRECTNESS OF THIS CODE OR ANY DERIVATIVE WORKS WHICH INCORPORATE
 * IT. AUTODESK PROVIDES THE CODE ON AN  "AS-IS" BASIS AND EXPLICITLY
 * DISCLAIMS ANY  LIABILITY, INCLUDING  CONSEQUENTIAL AND  INCIDENTAL
 * DAMAGES  FOR ERRORS,  OMISSIONS,  AND OTHER  PROBLEMS IN THE CODE.
 *
 * Use, duplication, or disclosure by the U.S. Government is subject
 * to restrictions set forth in  FAR 52.227-19  (Commercial Computer
 * Software  Restricted  Rights)  and  DFAR   252.227-7013(c)(1)(ii)
 * (Rights in Technical Data and Computer Software),  as applicable.
 *
 * @author: Valentin Koch <valentin.koch@autodesk.com>
 */
#ifndef LP_SOLVER_H_
#define LP_SOLVER_H_

#include <stdexcept>
#include <string>
#include <vector>
#include <limits>

#ifndef _DNT
#define _DNT(STRING) STRING
#endif

namespace aop
{

/**
 * @class LpSolver
 *
 * This abstract class represents an LP solver. It's purpose is to use different
 * brands of LP solvers for the same problem instance. This allows an easy switch
 * from one brand of solver to another.
 */
class LpSolver
{
public:
	/// Initializes a new, empty model.
	LpSolver() {};
	/// Destructor. Must be virtual.
	virtual ~LpSolver() {};

	/// Signs used in constraints.
	enum Sign { LESS_EQUAL, EQUAL, GREATER_EQUAL };
	/// Optimization sense.
	enum ModelSense { MINIMIZE, MAXIMIZE };
	/// The status of the solving process.
	enum Status { OPTIMAL, UNBOUNDED, INFEASIBLE, TIMEOUT, NUMERIC, UNKNOWN };

	/// Infinity.
	static constexpr double INFTY = std::numeric_limits<double>::infinity();

	/// Initializes a new model. Current model will be cleared.
	virtual void clear_model() = 0;

	/// Add continuous variable with lower bound lb, upper bound ub, and cost cost.
	virtual void add_var_continuous(const double lb, const double ub, const double cost, const std::string& name = "") = 0;

	/** 
	 * Add constraint with left hand side term given by variable positions in vars
	 * and corresponding coefficients in coef, sign, and right hand side constant rhs
	 */
	virtual void add_constraint(const std::vector<int>& vars, const std::vector<double>& coef,
				    const Sign sign, const double rhs, const std::string& name = "") = 0;

	/// Minimize or maximize (default for a model is minimize).
	virtual void set_model_sense(const ModelSense sense) = 0;

	/// Set a starting point (used either for warmstarts or for IPM's, if supported).
	virtual void set_starting_point(const std::vector<double>& s) {};

	/// Set the maximum solving time in seconds.
	virtual void set_maximum_solving_time(const double seconds) = 0;

	/// Solve the current model, return status after solve and fills current solution value and solution vector.
	virtual Status solve(double& value, std::vector<double>& solution) = 0;
};


/**
 * @class solver_error
 *
 * A runtime exception that communicates a general error that occurs in
 * the underlying LP solver.
 */
class solver_error : public std::runtime_error
{
public:
	/// Default Constructor.
	solver_error()
		: std::runtime_error(_DNT("An error in the LP solver occured.")) {};
	/// Constructor using STL-style argument.
	solver_error(std::string& m) : std::runtime_error(m) {};
	/// Constructor using C-style argument.
	solver_error(const char* m) : std::runtime_error(m) {};
	/// Destructor.
	virtual ~solver_error() throw() {};
};


/**
 * @class lp_unbounded_error
 *
 * This exception can be thrown if the LP is unbounded.
 */
class lp_unbounded_error : public solver_error
{
public:
	/// Default Constructor.
	lp_unbounded_error()
		: solver_error(_DNT("Problem is unbounded.")) {};
	/// Constructor using STL-style argument.
	lp_unbounded_error(std::string& m) : solver_error(m) {};
	/// Constructor using C-style argument.
	lp_unbounded_error(const char* m) : solver_error(m) {};
	/// Destructor.
	virtual ~lp_unbounded_error() throw() {};
};


/**
 * @class lp_infeasible_error
 *
 * This exception can be thrown if the LP is infeasible.
 */
class lp_infeasible_error : public solver_error
{
public:
	/// Default Constructor.
	lp_infeasible_error()
		: solver_error(_DNT("Problem is infeasible.")) {};
	/// Constructor using STL-style argument.
	lp_infeasible_error(std::string& m) : solver_error(m) {};
	/// Constructor using C-style argument.
	lp_infeasible_error(const char* m) : solver_error(m) {};
	/// Destructor.
	virtual ~lp_infeasible_error() throw() {};
};


/**
 * @class lp_timeout_error
 *
 * This exception can be thrown if the LP timed out.
 */
class lp_timeout_error : public solver_error
{
public:
	/// Default Constructor.
	lp_timeout_error()
		: solver_error(_DNT("Solve process timed out.")) {};
	/// Constructor using STL-style argument.
	lp_timeout_error(std::string& m) : solver_error(m) {};
	/// Constructor using C-style argument.
	lp_timeout_error(const char* m) : solver_error(m) {};
	/// Destructor.
	virtual ~lp_timeout_error() throw() {};
};


/**
 * @class lp_numeric_error
 *
 * This exception can be thrown if the solver encountered
 * numerical difficulities..
 */
class lp_numeric_error : public solver_error
{
public:
	/// Default Constructor.
	lp_numeric_error()
		: solver_error(_DNT("Aborted due to numeric difficulties.")) {};
	/// Constructor using STL-style argument.
	lp_numeric_error(std::string& m) : solver_error(m) {};
	/// Constructor using C-style argument.
	lp_numeric_error(const char* m) : solver_error(m) {};
	/// Destructor.
	virtual ~lp_numeric_error() throw() {};
};


} // namespace aop

#endif
